#!/usr/bin/perl -w
#
#  FORTRAN Object
#  Use this module to read and write FORTRAN files
#
#  Copyright (C) 2001  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with the GPSd programs; see the file COPYING.  If not, write
#  to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.

package FORTRAN;

use strict;
use FileHandle;
use IO::Seekable;

#
#  Open file
#
sub open
{
   my $class = shift;
   my $self = {};
   bless($self,$class);

   my ($file,$lahey) = @_;

   my $fh = new FileHandle($file);
   defined($fh) || return;
   binmode($fh);
   $self->{WR} = ($file =~ /^ *>/ || $file =~ /^ *|/);
   $self->{FH} = $fh;
   $self->{LF} = $lahey;
   if ($self->{LF})
   {
      (!$self->{WR}) || die "Writing LAHEY files not supported\n";
      my $buffer;
      (read($self->{FH},$buffer,1)==1) || die "Error reading file header\n";
      (unpack("C",$buffer) == 0xFD) || die "File is not a Lahey file\n";
   }
   return $self;
}

#
#  Rewind file
#
sub rewind
{
   my $self = shift;
   seek($self->{FH},0,SEEK_SET) || return;
   if ($self->{LF})
   {
      my $buffer;
      (read($self->{FH},$buffer,1)==1) || die "Error reading file header\n";
      (unpack("C",$buffer) == 0xFD) || die "File is not a Lahey file\n";
   }
   return 1;
}

#
#  Backspace file
#
sub backspace
{
   my $self = shift;
   my ($n) = @_;
   defined($n) || ($n = 1);
   ($n>0) || die "Invalid backspace count $n\n";
   for (;$n>0;$n--)
   {
      #  Backspace beyond BOF
      (tell($self->{FH})>1) || return;
      if ($self->{LF})
      {
         die "Cannot backspace on Lahey files\n";
      }
      else
      {
         seek($self->{FH},-4,SEEK_CUR) || return;
         my $buffer;
         (read($self->{FH},$buffer,4)==4) || die "Error reading file header\n";
         my $n = unpack("L",$buffer);
         seek($self->{FH},-8-$n,SEEK_CUR) || return;
      }
   }
   return 1;
}

#
#  Close file
#
sub close
{
   my $self = shift;
   close($self->{FH});
}

#
#  Read record head
#
sub head
{
   my $self = shift;
   my $n;
   if ($self->{LF})
   {
      my $buffer;
      (read($self->{FH},$buffer,1)==1) || return;
      my $k = unpack("C",$buffer);
      my $m = ($k&0x03);
      for (my $i=1;$i<=$m;$i++)
      {
         (read($self->{FH},$buffer,1)==1) || die "Error reading record header";
         $k += unpack("C",$buffer)<<(8*$i);
      }
      $n = int($k/4);
   }
   else
   {
      my $buffer;
      (read($self->{FH},$buffer,4)==4) || return;
      $n = unpack("L",$buffer);
   }
   return $n;
}

#
#  Read record tail
#
sub tail
{
   my $self = shift;
   my ($n) = @_;
   if ($self->{LF})
   {
      #  Record length including head and tail
      my $k = ($n<<2);
      if (0xFF000000 & $k)
      { $k += 8*4+3; }
      elsif (0xFFFF0000 & $k)
      { $k += 4*4+2; }
      elsif (0xFFFFFF00 & $k)
      { $k += 2*4+1; }
      else
      { $k += 1*4+0; }
      #  Tail length
      my $l=0;
      for (my $m=$k;$m>0;$m = ($m>>8))
      {
         my $buffer;
         (read($self->{FH},$buffer,1)==1) || die "Error reading record tail";
         $l = ($l<<8) + unpack("C",$buffer);
      }
#     printf "n=%d k=%d l=%d l3=%d k3=%d l/4=%d k/4=%d\n" , $n , $k , $l , ($l&3) , ($k&3) , int($l/4) , int($k/4);
#    ($l == $k) || die "Head and tail mismatch\n";
   }
   else
   {
      my $buffer;
      (read($self->{FH},$buffer,4)==4) || die "Error reading record tail\n";
      my $l = unpack("L",$buffer);
      ($l == $n) || die "Head and tail mismatch $l != $n\n";
   }
}

#
#  Read record
#
sub read
{
   my $self = shift;
   my $buffer;
   my $n = $self->head();
   ($n) || return;
   (read($self->{FH},$buffer,$n)==$n) || die "Error reading file\n";
   $self->tail($n);
   return $buffer;
}

#
#  Skip record
#
sub skip
{
   my $self = shift;
   my ($n) = @_;
   defined($n) || ($n = 1);
   ($n>0) || die "Invalid skip count $n\n";
   for (;$n>0;$n--)
   {
      my $k = $self->head();
      ($k) || return;
      seek($self->{FH},$k,SEEK_CUR) || die "Error seeking file\n";
      $self->tail($k);
   }
   return 1;
}

#
#  Write record
#
sub write
{
   my $self = shift;
   my ($buffer) = @_;
   my $fh = $self->{FH};
   return print $fh pack("La*L",length($buffer),$buffer,length($buffer));
}

1;
